#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <glade/glade.h>

#include "str_util.h"
#include "elist.h"
#include "dir_util.h"
#include "mru.h"
#include "prefs.h"
#include "main_win.h"
#include "cursor.h"
#include "edit_tab.h"

#include "dir_select.h"


/* widgets we need */
static GtkCombo *combo_wd = NULL;
static GtkEntry *ent_wd = NULL;
static GtkToggleButton *cb_recurse = NULL;
static GtkFileSelection *dlg_wd_select = NULL;

/* private data */
static GString *working_dir = NULL;
static GEList *file_list = NULL;

/* preferences */
static gboolean *recurse;
static MRUList *dir_mru;


/*** private functions ******************************************************/

static gboolean check_dir(GString *dir)
{
	int res;
	struct stat stat_data;

	if (strlen(dir->str) == 0) {
		g_string_assign(dir, g_get_home_dir());
	}

	res = stat(dir->str, &stat_data);
	if (res < 0) {
		mw_status_printf("Error: Can't open directory \"%s\".", dir->str);
		return FALSE;
	}
	if (!S_ISDIR(stat_data.st_mode)) {
		mw_status_printf("Error: \"%s\" is not a directory.", dir->str);
		return FALSE;
	}
	if (!du_check_permission(&stat_data, "rx")) {
		mw_status_printf("Error: Permission denied for \"%s\".", dir->str);
		return FALSE;
	}

	if (dir->str[strlen(dir->str)-1] != '/')
		g_string_append_c(dir, '/');
	return TRUE;
}


/* UI callbacks */
void cb_select_dir(GtkWidget *widget, gpointer data)
{
	gtk_file_selection_set_filename(dlg_wd_select, working_dir->str);
	gtk_widget_show(GTK_WIDGET(dlg_wd_select));
}

void cb_set_working_dir(GtkWidget *widget, gpointer data)
{
	ds_set_working_dir(gtk_file_selection_get_filename(dlg_wd_select));
}

void cb_edit_dir(GtkWidget *widget, gpointer data)
{
	char *new_dir;

	new_dir = strdup(gtk_entry_get_text(GTK_ENTRY(widget)));
	ds_set_working_dir(new_dir);
	free(new_dir);
}

gboolean cb_keypress_dir(GtkWidget *widget, GdkEventKey *event, gpointer user_data)
{
	char *new_dir;

	if (event->keyval == GDK_Return) {
		/* stop the combo box from displaying the dropdown list */
		gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "key_press_event");

		new_dir = strdup(gtk_entry_get_text(GTK_ENTRY(widget)));
		if (strcmp(working_dir->str, new_dir) == 0)
			ds_set_working_dir(NULL);
		else
			ds_set_working_dir(new_dir);
		free(new_dir);

		return TRUE;
	}

	return FALSE;
}

void cb_toggle_recurse(GtkWidget *widget, gpointer data)
{
	*recurse = gtk_toggle_button_get_active(cb_recurse);
	ds_set_working_dir(NULL);
}


/*** public functions *******************************************************/

void ds_init(GladeXML *xml)
{
	GEList *dir_list;

	/*
	 * get the widgets from glade
	 */

	combo_wd = GTK_COMBO(glade_xml_get_widget(xml, "combo_wd"));
	ent_wd = GTK_ENTRY(glade_xml_get_widget(xml, "ent_wd"));
	cb_recurse = GTK_TOGGLE_BUTTON(glade_xml_get_widget(xml, "cb_recurse"));
	dlg_wd_select = GTK_FILE_SELECTION(glade_xml_get_widget(xml, "dlg_wd_select"));

	gtk_file_selection_hide_fileop_buttons(dlg_wd_select);
	gtk_widget_set_sensitive(dlg_wd_select->file_list, FALSE);

	/*
	 * get the preference values, or set them to defaults
	 */

	/* recurse */
	recurse = pref_get_ref("ds:recurse");
	if (!recurse) {
		gboolean temp = FALSE;
		recurse = pref_set("ds:recurse", PREF_BOOLEAN, &temp);
	}

	gtk_toggle_button_set_active(cb_recurse, *recurse);

	/* dir_mru */
	dir_list = pref_get_ref("ds:dir_mru");
	if (!dir_list) {
		GEList *temp_list = g_elist_new();
		g_elist_append(temp_list, g_get_home_dir());
		dir_list = pref_set("ds:dir_mru", PREF_STRING | PREF_LIST, temp_list);
		g_elist_free_data(temp_list);
	}
	dir_mru = mru_new_from_list(10, dir_list);
}


void ds_set_initial_dir(gchar *dir)
{
	working_dir = g_string_sized_new(200);

	if (dir == NULL)
		ds_set_working_dir(g_elist_first(dir_mru->list)->data);
	else {
		if (g_path_is_absolute(dir))
			ds_set_working_dir(dir);
		else {
			/* make the path absolute */
			char *cur = g_get_current_dir();
			char *abs = du_join_path(cur, dir);
			ds_set_working_dir(abs);
			free(cur);
			free(abs);
		}
	}
}


void ds_set_working_dir(gchar *dir)
{
	static const char *patterns[] = {
#ifdef ENABLE_MP3
		"*.[mM][pP][aA23]",
#endif
#ifdef ENABLE_VORBIS
		"*.[oO][gG][gG]",
#endif
		NULL
	};
	gint res, count;

	if (dir != NULL)
		if (strcmp(working_dir->str, dir) == 0)
			return;

	/* update the directory mru list */
	if (dir != NULL) {
		mru_add(dir_mru, dir);
		gtk_combo_set_popdown_strings(combo_wd, GLIST(dir_mru->list));
	}

	/* clear the form in the edit tab */ 
	et_unload_file();

	mw_status_printf("Scanning...");
	cursor_set_wait();

	/* flush pending operations before we start */
	while (gtk_events_pending()) gtk_main_iteration();

	/* free the current file_list */
	if (file_list) {
		g_elist_free_data(file_list);
		file_list = NULL;
	}

	/* set the new working_dir and file_list */
	if (dir != NULL) {
		g_string_assign(working_dir, dir);
		res = check_dir(working_dir);
		gtk_entry_set_text(ent_wd, working_dir->str);
		if (!res) {
			file_list = g_elist_new();
			et_update_file_list(NULL);
			return;
		}
	}
	file_list = du_get_file_list(working_dir->str, patterns, 
			gtk_toggle_button_get_active(cb_recurse),
			TRUE);

	/* update the interface */
	et_update_file_list(file_list);
	count = g_elist_length(file_list);
	switch (count) {
	    case 0:
		mw_status_printf("No files found.");
		break;
	    case 1:
		mw_status_printf("1 file found.");
		break;
	    default:
		mw_status_printf("%d files found.", count);
		break;
	}

	cursor_set_normal();
}


const gchar *ds_get_working_dir()
{
	return working_dir->str;
}


GEList *ds_get_file_list()
{
	return file_list;
}


