#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <langinfo.h>
#include <iconv.h>
#include <errno.h>
#include <math.h>
#include <glib.h>

#include "str_util.h"
#include "math_util.h"


/*** private functions ******************************************************/

static const char *encoding_name(int enc)
{
	static char *names[] = {
		"UTF-8",
		"UTF-16",
		"UTF-16BE",
		"UTF-16LE",
		"ISO-8859-1"
	};

	if (enc == LOCALE)
		return nl_langinfo(CODESET);
	else
		return names[enc];
}


/*
 * Returns the (minimum) number of bytes per character in this encoding.
 *
 * XXX - Don't know a generic way to determine bytes per character for all 
 *       encodings.  For now multibyte encodings other than UTF-16 will be wrong.
 */
static int bytes_per_char(int enc)
{
	if (strstr(encoding_name(enc), "UTF-16") != NULL)
		return 2;
	else
		return 1;
}


/*
 * Determines the size of a string in bytes, given the encoding.
 * Size includes any markers (Unicode BOM, terminating zero(s), etc.)
 *
 * XXX - Result will be wrong in those cases where bytes_per_char() is wrong.
 */
static size_t strsize(int enc, const void *str)
{
	const void *p = str;
	int bpc = bytes_per_char(enc);

	if (bpc != 1 && bpc != 2 && bpc != 4) {
		fprintf(stderr, "strsize: can't handle %i bytes per char\n", bpc);
		return 0;
	}

	while(1) {
		if ( (bpc == 1 && *(guint8 *)p == 0) ||
		     (bpc == 2 && *(guint16*)p == 0) ||
		     (bpc == 4 && *(guint32*)p == 0) )
			break;
		p += bpc;
	}

	return p - str + bpc;
}


/*** public functions *******************************************************/

char *str_convert_encoding(int from, int to, const char *str)
{
	iconv_t conv;
	char *result;
	char *inbuf, *outbuf;
	size_t inbpc, outbpc;
	size_t inbytes, outbytes;
	size_t inbytesleft, outbytesleft;
	size_t res;

	inbytes = strsize(from, str);

	if (from == to) {
		result = malloc(inbytes);
		memcpy(result, str, inbytes);
		return result;
	}

	conv = iconv_open(encoding_name(to), encoding_name(from));
	if (iconv == (iconv_t)-1) {
		fprintf(stderr, "convert_encoding: cannot convert from %s to %s\n", 
			encoding_name(from), encoding_name(to));
		return NULL;
	}

	inbpc = bytes_per_char(from);
	outbpc = bytes_per_char(to);

	/* guesstimate the converted size (tuned for european languages...) */
	outbytes = ((double)outbpc / (double)inbpc) * inbytes;
	if (to == UTF_8)
		outbytes = ceil(1.2 * inbytes);
	else if (to == UTF_16)
		outbytes += 2;	/* for the BOM */

	//printf("inbytes : %i\noutbytes: %i\n", inbytes, outbytes);

	result = malloc(outbytes);

	inbuf = (char*)str;
	inbytesleft = inbytes;
	outbuf = result;
	outbytesleft = outbytes;

	while(1) {
		res = iconv(conv, &inbuf, &inbytesleft, &outbuf, &outbytesleft);
		if (res == (size_t)-1) {
			if (errno == E2BIG) {
				/* Ran out of space, alloc more 
				   This code tries hard to avoid the need for a second realloc,
				   while still keeping over-allocation to a minimum */
				double done = 1.0 - (double)inbytesleft / (double)inbytes;
				size_t bytes_written = outbuf - result;
				size_t newsize = ceil((bytes_written / done) * 1.1);

				//printf("growing: done=%g%%, old size=%i, new size=%i\n",
				//	100.0*done, outbytes, newsize);

				outbytesleft += newsize - outbytes;
				outbytes = newsize;
				result = realloc(result, outbytes);
				outbuf = result + bytes_written;

				continue;
			}
			else {
				/* Invalid or inconvertible char, skip it
				   Seems better than aborting the conversion... */

				fprintf(stderr, "convert_encoding: conversion error at offset %i\n", 
					inbytes-inbytesleft);

				inbuf += inbpc;
				inbytesleft = max(inbytesleft - inbpc, 0);
				outbuf += outbpc;
				outbytesleft = max(outbytesleft - outbpc, 0);

				continue;
			}
		}
		break;
	}

	//printf("%i of %i bytes unused (wasted %g%%)\n", 
	//	outbytesleft, outbytes, 100.0*(double)outbytesleft/(double)outbytes);

	iconv_close(conv);
	return result;
}


void str_convert_case(char *str, int conv)
{
	switch (conv) {
		case CASE_CONV_NONE: {
			return;
		}
		case CASE_CONV_LOWER: {
			char *p;
			for (p = str; *p ; p++)
				*p = tolower(*p);
			break;
		}
		case CASE_CONV_UPPER: {
			char *p;
			for (p = str; *p ; p++)
				*p = toupper(*p);
			break;
		}
		case CASE_CONV_SENTENCE: {
			char *p;
			for (p = str; *p ; p++) {
				if (p == str)
					*p = toupper(*p);
				else
					*p = tolower(*p);
			}
			break;
		}
		case CASE_CONV_TITLE: {
			char *p;
			int upcase_next = 1;
			for (p = str; *p ; p++) {
				if (upcase_next)
					*p = toupper(*p);
				else
					*p = tolower(*p);
				upcase_next = !isalnum(*p) && *p != '\'';
			}
			break;
		}
		default: {
			fprintf(stderr, "str_convert_case: unrecognized convertion type (%i), ignoring\n", conv);
			break;
		}
	}
}


void str_trim(char *str, char c)
{
	int i;

	for (i = strlen(str); --i>=0 && str[i]==c; );
	str[i+1] = 0;
}


void str_replace_char(char *str, char r, char c)
{
	char *p;
	for (p = str; *p; p++)
		if (*p == r)
			*p = c;
}


void str_replace_chars(char *str, const char *r, char c)
{
	char *p;
	for (p = str; *p; p++)
		if (strchr(r, *p))
			*p = c;
}

void str_remove_char(char *str, char rem)
{
	char *r = str;
	char *w = r;
	while (*r) {
		if (*r != rem)
			w++;
		r++;

		if (r != w)
			*w = *r;
	}
}

char *str_strnchr(const char *s, char c, int n)
{
	int p = (int)s;
	int count = 0;

	while (*(char *)p) {
		if (*(char *)p == c)
			if (++count == n)
				return (char *)p;
		p++;
	}
	return NULL;
}

char *str_strnrchr(const char *s, char c, int n)
{
	int p = (int)s + strlen(s);
	int count = 0;

	while (p >= (int)s) {
		if (*(char *)p == c)
			if (++count == n)
				return (char *)p;
		p--;
	}
	return NULL;
}

char *str_safe_strncpy(char *dest, const char *src, size_t n)
{
	strncpy(dest, src, n);
	dest[n-1] = 0;
	return dest;
}


